;;; $Id: arith-expr.scm,v 1.6 1999/02/21 22:40:33 leavens Exp $

;;; Helping procedures for the following grammar.
;;;    <arith-expr> ::= <number>
;;;                    | (<arith-expr> <op> <arith-expr>)
;;;                    | (- <arith-expr>)
;;;   <op> ::= + | - | * | **

(define-record literal (number))
(define-record op-call (left-arg op right-arg))
(define-record unary-minus (arg))

(define arith-expr:permitted-ops
  ;; TYPE: (list symbol)
  '(+ - * **))

(define arith-expr?
  ;; TYPE: (-> (datum) boolean)
  (lambda (x)
    (or (and (literal? x) (number? (literal->number x)))
	(and (op-call? x)
	     (arith-expr? (op-call->left-arg x))
	     (arith-expr? (op-call->right-arg x))
	     (memq (op-call->op x) arith-expr:permitted-ops))
	(and (unary-minus? x)
	     (arith-expr? (unary-minus->arg x))))))

(define parse-arith-expr
  ;; TYPE: (-> (datum) arith-expr)
  (lambda (exp)
    (cond
     ((number? exp) (make-literal exp))
     ((and (list? exp) (= (length exp) 2) (eq? (car exp) '-))
      (make-unary-minus (parse-arith-expr (cadr exp))))
     ((and (list? exp)
	   (= (length exp) 3)
	   (memq (cadr exp) arith-expr:permitted-ops))
      (make-op-call (parse-arith-expr (car exp))
		     (cadr exp)
		     (parse-arith-expr (caddr exp))))
     (else (error "parse-arith-expr: bad syntax: " exp)))))

(define unparse-arith-expr
  ;; TYPE: (-> (arith-expr) datum)
  (lambda (exp)
    (variant-case exp
     (literal (number) number)
     (op-call (left-arg op right-arg)
       (list (unparse-arith-expr left-arg)
	     op
	     (unparse-arith-expr right-arg)))
     (unary-minus (arg)
       (list '- (unparse-arith-expr arg)))
     (else (error "unparse-arith-expr: bad argument: " exp)))))
