;;; $Id: atomic-exp.scm,v 1.1 1999/02/15 23:03:25 leavens Exp $
;;; "atomic-exps", a form of atomic-expressions, using the grammar
;;;    <(atomic-exp T)> ::= <T> | ( {<(atomic-exp T)>}* )
;;; where the T's are required to be atomic and not null.

(load-quietly-from-lib "all.scm")
(load-quietly-from-lib "atomic.scm")

(define atomic-list?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (and (list? d)
	 (all atomic-exp? d))))

(define atomic-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (or (atomic? d)
	(atomic-list? d))))

(define atom->atomic-exp
  ;; TYPE: (-> (T) (atomic-exp T))
  (lambda (t)
    (if (and (atomic? t) (not (null? t)))
	t
	(error
	 "atom->atomic-exp: argument must be atomic and not null, but is: "
	 t))))

(define atomic-list->atomic-exp
  ;; TYPE: (-> ((list (atomic-exp T))) (atomic-exp T))
  (lambda (tr)
    tr))

(define atomic-exp-atomic?
  ;; TYPE: (-> ((atomic-exp T)) boolean)
  (lambda (exp)
    (and (atomic? exp) (not (null? exp)))))

(define atomic-exp-atomic-list?
  ;; TYPE: (-> ((atomic-exp T)) boolean)
  (lambda (exp)
    (or (null? exp) (pair? exp))))

(define atomic-exp->atom
  ;; TYPE: (-> ((atomic-exp T)) T)
  (lambda (exp)
    (if (atomic-exp-atomic? exp)
	exp
	(error "atomic-exp->atom: this is not an atom: " exp))))

(define atomic-exp->atomic-list
  ;; TYPE: (-> ((atomic-exp T)) (list (atomic-exp T)))
  (lambda (exp)
    (if (atomic-exp-atomic-list? exp)
	exp
	(error "atomic-exp->atom: this is not an atomic-list: " exp))))

(define parse-atomic-exp
  ;; TYPE: (and (-> ((list T)) (atomic-exp T))
  ;;		(-> (U) (atomic-exp U))
  ;;		(-> (datum) (atomic-exp datum)))
  (lambda (exp)
    (cond
     ((list? exp) (atomic-list->atomic-exp (parse-atomic-list exp)))
     ((atomic? exp) (atom->atomic-exp exp))
     (else (error "parse-atomic-exp: bad syntax: " exp)))))

(define parse-atomic-list
  ;; TYPE: (and (-> ((list T)) (list (atomic-exp T)))
  ;;            (-> ((list datum)) (list (atomic-exp datum)))))
  (lambda (exp)
    (cond
     ((list? exp) (map parse-atomic-exp exp))
     (else (error "parse-atomic-list: bad syntax: " exp)))))
