;;; $Id: atomic-expression.scm,v 1.6 1998/09/28 20:32:02 leavens Exp $
;;; Atomic-Expressions, using the grammar:
;;;    <(atomic-expression T)> ::= <T> | <(tree T)>
;;; where the T's are required to be atomic and not null.
;;; See the file atomic-tree.scm for the grammar and operations on trees.
;;; An equivalent grammar, sometimes used to prevent mutual recursion, is:
;;;    <(atomic-expression T)> ::= <T> | ( {<(atomic-expression T)>}* )

(load-quietly-from-lib "all.scm")
(load-quietly-from-lib "atomic.scm")

(define atomic-expression?
   ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (or (atomic? d)
	(and (list? d)
	     (all atomic-expression? d)))))
 
(define atom->atomic-expression
  ;; TYPE: (-> (T) (atomic-expression T))
  (lambda (t)
    (if (and (atomic? t) (not (null? t)))
	t
	(error
	 "atom->atomic-expression: argument must be atomic and not null, but is: "
	 t))))

(define tree->atomic-expression
  ;; TYPE: (-> ((tree T)) (atomic-expression T))
  (lambda (tr)
    (if (and (list? tr) (atomic-expression? tr))
	tr
	(error "list->atomic-expression: improper tree argument: " tr))))

(define atomic-expression-atom?
  ;; TYPE: (-> ((atomic-expression T)) boolean)
    (lambda (exp)
      (and (atomic? exp) (not (null? exp)))))
      
(define atomic-expression-tree?
  ;; TYPE: (-> ((atomic-expression T)) boolean)
  (lambda (exp)
    (or (null? exp) (pair? exp))))

(define atomic-expression->atom
  ;; TYPE: (-> ((atomic-expression T)) T)
  (lambda (exp)
    (if (atomic-expression-atom? exp)
	exp
	(error "atomic-expression->atom: atomic-expression is a tree " exp))))

(define atomic-expression->tree
  ;; TYPE: (-> ((atomic-expression T)) (tree T))
  (lambda (exp)
    (if (atomic-expression-tree? exp)
	exp
	(error "atomic-expression->atom: atomic-expression is not a tree "
	       exp))))

(define parse-atomic-expression
  ;; TYPE: (and (-> ((list T)) (atomic-expression T))
  ;;            (-> (S) (atomic-expression S))
  ;;            (-> (datum) (atomic-expression datum))))
  (lambda (exp)
    (cond
     ((list? exp) (map parse-atomic-expression exp))
     ((atomic? exp) (atom->atomic-expression exp))
     (else (error "parse-atomic-expression: bad syntax: " exp)))))
