;;; $Id: atomic-tree.scm,v 1.9 1999/04/29 15:35:29 leavens Exp $
;;; Trees of atomic items, using the grammar:
;;;    <(tree T)> ::= ( {<(atomic-expression T)>}* )
;;;    <(atomic-expression T)> ::= <T> | <(tree T)>
;;; where the T's are required to be atomic and not null.
;;; This is equivalent to:
;;;    <(tree T)> ::= () 
;;;                |  ( <T> . <(tree T)> )
;;;                |  ( <(tree T)> . <(tree T)> )

(load-quietly-from-lib "atomic-expression.scm")

(define tree?
  ;; TYPE: (-> (datum) boolean)
  (lambda (x)
    (and (list? x)
	 (all atomic-expression? x))))
		  
(define tree
  ;; TYPE: (-> ((atomic-expression T)...) (tree T))
  (lambda aes
    aes))

(define tree-empty
  ;; TYPE: (-> () (tree T))
  (lambda ()
    '()))

(define tree-add
  ;; TYPE: (-> ((atomic-expression T) (tree T)) (tree T))
  cons)

(define tree-add-atom
  ;; TYPE: (-> (T (tree T)) (tree T))
  (lambda (e tr)
    (tree-add (atom->atomic-expression e) tr)))

(define tree-add-tree
  ;; TYPE: (-> ((tree T) (tree T)) (tree T))
  (lambda (e tr)
    (tree-add (tree->atomic-expression e) tr)))

(define tree-null?
  ;; TYPE: (-> (datum) boolean)
  null?)

(define tree-first-atomic?
  ;; TYPE: (-> ((tree T)) boolean)
  (lambda (tr)
    (atomic-expression-atom? (car tr))))

(define tree-first-tree?
  ;; TYPE: (-> ((tree T)) boolean)
  (lambda (tr)
    (atomic-expression-tree? (car tr))))

(define tree-first
  ;; TYPE: (-> ((tree T)) (atomic-expression T))
  car)

(define tree-first-atom
  ;; TYPE: (-> ((tree T)) T)
  (lambda (tr)
    (atomic-expression->atom (tree-first tr))))

(define tree-first-tree
  ;; TYPE: (-> ((tree T)) (tree T))
  (lambda (tr)
    (atomic-expression->tree (tree-first tr))))

(define tree-rest
  ;; TYPE: (-> ((tree T)) (tree T))
  cdr)

(define parse-tree
  ;; TYPE: (-> (datum) (tree datum))
  (lambda (d)
    (if (list? d)
	(map parse-atomic-expression d)
	(error "parse-tree: bad syntax: " d))))
