;; $Id: ch6-2.scm,v 1.11 1998/11/12 22:53:20 leavens Exp $

;; Code from the EOPL chapter 6 revision, figure 6.2.1
;; (call by reference with arrays in the indirect model).

;;; Grammar:
;;          see the file ch6-base.scm for details,
;;          except that the production for <operand> is replaced by
;;                <operand> ::= <varref>
;;                            | <array-exp> [<exp>]
;;                            | <exp>

(load-from-lib "ch6-base.scm")

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = Cell(Expressed-Value) + Array-Element(Expressed-Value)
;;; Expressed-Value = Number + Procedure + Array(Expressed-Value)
;;;                   + List(Expressed-Value) + Void
;;; Procedure = prim-proc + closure
;;; Array(T) = {Cell(T)}*

(load-quietly-from-lib "ch6-2-denoted-value.scm")
(load-quietly-from-lib "ch6-1-3-expressed-value.scm")


;;; *Modified* from Figure 6.2.1, page 191 and Figure 6.1.3, page 183

(define denoted->expressed
  ;; TYPE: (-> (Denoted-Value) Expressed-Value)
  (lambda (den-val)
    (cond
      ((cell? den-val) (cell-ref den-val))
      ((ae? den-val) (array-ref (ae->array den-val) (ae->index den-val)))
      (else (error "Can't dereference denoted value: " den-val)))))

(define expressed->denoted
  ;; TYPE: (-> (Expressed-Value) Denoted-Value)
  make-cell)

(define denoted-value-assign!
  ;; TYPE: (-> (Denoted-Value Expressed-Value) void)
  (lambda (den-val val)
    (cond
      ((cell? den-val) (cell-set! den-val val))
      ((ae? den-val)
       (array-element-assign!  ; changed to better support direct model!
	(ae->array den-val)
	(ae->index den-val)
	val))
      (else (error "Can't assign to denoted value: " den-val)))))

(define do-letarray
  ;; TYPE: (-> (number) Denoted-Value)
  (lambda (n)
    (if (not (<= 0 n))  ; added error check vs. text
	(error "do-letarray: bad index")
	(make-cell (array->expressed (make-array n))))))

(define eval-array-exp
  ;; TYPE: (-> (parsed-exp Environment) Expressed-Value)
  (lambda (exp env)
    (eval-exp exp env)))

;; addition to figure 6.1.3

(define array-element-assign!
  ;; TYPE: (-> ((Array Expressed-Value) number Expressed-Value) void)
  array-set!)

;;; *Modified* from Figure 6.2.1 : page 191

(define eval-rand
  ;; TYPE: (-> (parsed-exp Environment) Denoted-Value)
  (lambda (rand env)
    (variant-case rand
      (varref (var) (apply-env env var))
      (arrayref (array index)
        (make-ae (expressed->array (eval-array-exp array env))
		 (expressed->number (eval-exp index env))))
      (else (expressed->denoted (eval-exp rand env))))))   ;; changed!
