;; $Id: ch7-2.scm,v 1.18 1999/08/11 16:27:59 leavens Exp $

;; Code from EOPL chapter 7, figure 7.2.4
;; (the inheritance language).

;;; Grammar:
;;         <form> ::= define <var> = <exp> | <exp>
;;	    <exp> ::= <integer-literal> | <varref>
;;		    | <instance-var> | <class-var>
;;                  | <instance-var> := <exp>
;;                  | <class-var> := <exp>
;;                  | method <varlist> <exp>
;;                  | $ <var> (<exps>)
;;                  | simpleinstance <exp>
;;                  | simpleclass <varlist> <varlist> <methdecls> <exp>
;;                  | class <exp>, <varlist> <varlist> <methdecls> <exp>
;;                  | $ <var> <super-rands>
;;		    | <operator> <operands>
;;                  | if <exp> then <exp> else <exp>
;;	            | proc <varlist> <exp>
;;	            | <var> := <exp>
;;                  | begin <exp> {; <exp>}* end
;;  <instance-var> ::= & <var>
;;     <class-var> ::= && <var>
;;    <meth-decls> ::= ( ) | ( <decls> )
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( ) | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;   <super-rands> ::= (super {, <exp>}*)
;;	  <varref> ::= <var>
;;	   <decls> ::= <decl> {;<decl>}*
;;	    <decl> ::= <var> = <exp>
;;	 <varlist> ::= ( ) | ( <vars> )
;;	    <vars> ::= <var> {,<var>}*

;;; Parsing

;; To use "run", you also need to do
;;        (load "run.scm")
;; To use "read-eval-print", you also need to do
;;        (load "rep-for-oop.scm")
;; Note that you can't use both of these at the same time.

(load-quietly-from-lib "undefine-read-eval-print.scm")
(load-quietly-from-lib "undefine-run.scm")
(load-quietly-from-lib "ch7-char-parser.scm")

;;; Syntactic sugars
;;       see the file below

(load-quietly-from-lib "syntax-expand-OOP.scm")

;;; Top-level Forms

(load-quietly-from-lib "eval-form-for-oop.scm")

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = Cell(Expressed-Value)
;;; Expressed-Value = Number + Procedure + List(Expressed-Value) + Void
;;;                   + Instance + Class + Open-Method
;;; Procedure = prim-proc + closure
;;; Instance = Class x Vector(Denoted-Value)
;;; Class = Maybe(Class) x {<var>}* x Vector(Denoted-Value) x {<var>}*
;;;         x Method-Environment
;;; Open-Method = Class-Thunk -> Method
;;; Class-Thunk = () -> Class
;;; Method = {Expressed-Value}* -> Expressed-Value
;;; Method-Environment = <var> -> Method

;;; ADTs used

(load-quietly-from-lib "environments.scm")
(load-quietly-from-lib "method-environment.scm")
(load-quietly-from-lib "procedure.scm")
(load-quietly-from-lib "maybe.scm")

;;; Helpful stuff

(load-quietly-from-lib "compose.scm")
(load-quietly-from-lib "ignore.scm")

;;; Truth values

;; Anonymous figure : page 146

(define true-value?
  ; TYPE: (-> (integer) boolean)
  (lambda (x)
    (not (zero? x))))

(define boolean->int
  ; TYPE: (-> (boolean) integer)
  (lambda (x)
    (if  x  1  0)))

;;; The following have to be after all the specific domains are defined.
(load-quietly-from-lib "ch7-1-denoted-value.scm")
(load-quietly-from-lib "ch7-2-expressed-value.scm")



;;; Evaluation of expressions

;;; Modified from Figure 7.2.4 : page 228

(define eval-exp
  ;; TYPE: (-> (parsed-exp Environment (maybe Class) (maybe Instance))
  ;;           Expressed-Value)
  (lambda (exp env class inst)
    (variant-case exp
      (new-class (parent-exp c-vars i-vars methdecls init-exp)
	(let ((parent-class (eval-exp parent-exp env class inst))
	      (open-methods
	       (map (lambda (decl)
		      (expressed->open-method
		       (eval-exp (decl->exp decl) env class inst)))
		    methdecls)))
	  (let ((new-c-vars (append c-vars (class->c-vars parent-class)))
		(new-i-vars (append i-vars (class->i-vars parent-class)))
		(new-c-vals ; NEW!
		 (make-shared-c-vals (class->c-vals parent-class) c-vars)))
	    (letrec ((new-class
		      (make-class
		       (make-something parent-class)
		       new-c-vars
		       new-c-vals
		       new-i-vars
		       (extend-method-env
			(map decl->var methdecls)
			(map (lambda (open-method)
			       (open-method (lambda () new-class)))
			     open-methods)
			(class->m-env parent-class)))))
	      (ignore
	       (eval-exp init-exp
			 env
			 (make-something new-class)
			 (make-nothing)))
	      new-class))))
      (super-meth-app (name rands)
	(let ((args (map (lambda (x) (eval-exp x env class inst)) rands)))
	  (meth-call name
		     (something->value
		      (class->parent
		       (something->value class)))
		     (cons (something->value inst) args))))
      ;; Figure 7.1.3 : page 219
      (i-varref (var)
	(lookup var
		(class->i-vars (something->value class))
		(instance->i-vals (something->value inst))))
      (c-varref (var)
	(lookup var
		(class->c-vars (something->value class))
		(class->c-vals (something->value class))))
      (i-varassign (var exp)
        (let ((value (eval-exp exp env class inst)))
          (void->expressed
	   (assign! var
		    value
		    (class->i-vars (something->value class))
		    (instance->i-vals (something->value inst))))))
      (c-varassign (var exp)
        (let ((value (eval-exp exp env class inst))
              (c-vals (class->c-vals (something->value class))))
          (void->expressed
	   (assign! var
		    value
		    (class->c-vars (something->value class))
		    c-vals))))
      (method (formals body)
        (let ((new-formals (cons 'self formals)))
	  (open-method->expressed
	   (lambda (class-thunk)
	     (lambda (args)
	       (eval-exp
		body
		(extend-env new-formals (map expressed->denoted args) env)
		(make-something (class-thunk))
		(make-something (car args))))))))
      (meth-app (name rands)
        (let ((args (map (lambda (rand)
			   (eval-exp rand env class inst))
			 rands)))
          (meth-call name (instance->class (car args)) args)))
      (new-simpleinst (class-exp)
        (let ((inst-class (eval-exp class-exp env class inst)))
          (let ((new-inst (make-instance inst-class
                            (make-vals (class->i-vars inst-class)))))
            (ignore (meth-call 'initialize inst-class (list new-inst)))
            new-inst)))
      ;; The following is a convenience to make old examples work (stackclass).
      ;; Syntax-expand can't handle simpleclass, because then it would be
      ;; expanded away in section 7.1
      (new-simpleclass (c-vars i-vars methdecls init-exp)
	(eval-exp (make-new-class (make-varref 'baseobject)
				  c-vars i-vars methdecls init-exp)
		  env class inst))
      ;; Modified from Figure 6.1.2 : page 182
      (lit (datum) (number->expressed datum))
      (varref (var) (denoted->expressed (apply-env env var)))
      (app (rator rands) 
	(apply-proc (expressed->procedure (eval-exp rator env class inst))
		    (map (lambda (rand)
			   (eval-rand rand env class inst))
			 rands)
		    class inst))
      (if (test-exp then-exp else-exp)
	(if (true-value?
	     (expressed->number (eval-exp test-exp env class inst)))
	    (eval-exp then-exp env class inst)
	    (eval-exp else-exp env class inst)))
      (proc (formals body) 
	(procedure->expressed (make-closure formals body env)))
      (varassign (var exp)
	(void->expressed
	 (denoted-value-assign! (apply-env env var)
				(eval-exp exp env class inst))))
      (begin (exp1 exp2)
	(begin
	  (ignore (eval-exp exp1 env class inst))
	  (eval-exp exp2 env class inst)))
      (else (error "Invalid abstract syntax: " exp)))))

;;; Modified from sources.ss for Figure 7.2.4

(define make-shared-c-vals
  ; TYPE: (-> ((vector Denoted-Value) (list symbol)) (vector Denoted-Value))
  (lambda (parent-c-vals c-vars)
    (list->vector
      (append
        (vector->list parent-c-vals)
        (map (lambda (x) (expressed->denoted (number->expressed 0)))
	     c-vars)))))

;;; From Figure 6.1.2 : p. 182

(define eval-rand
  ;; TYPE: (-> (parsed-exp Environment (maybe Class) (maybe Instance))
  ;;           Denoted-Value)
  (lambda (exp env class inst)
    (expressed->denoted (eval-exp exp env class inst))))

(define apply-proc
  ;; TYPE: (-> (Procedure (list Denoted-Value) (maybe Class) (maybe Instance))
  ;;           Expressed-Value)
  (lambda (proc args class inst)
    (variant-case proc
      (prim-proc (prim-op)
		 (apply-prim-op prim-op (map denoted->expressed args)))
      (closure (formals body env)
	       (eval-exp body (extend-env formals args env) class inst))
      (else (error "Invalid procedure:" proc)))))

;;; Figure 6.1.3 : page 183

(define denoted->expressed
  ; TYPE: (-> (Denoted-Value) Expressed-Value)
  cell-ref)

(define denoted-value-assign!
  ; TYPE: (-> (Denoted-Value Expressed-Value) void)
  cell-set!)

(define expressed->denoted
  ; TYPE: (-> (Expressed-Value) Denoted-Value)
  make-cell)

;;; Figure 7.1.4 : page 220

(define lookup
  ;; TYPE: (-> (symbol (list symbol) (vector Denoted-Value)) Expressed-Value)
  (lambda (var vars vals)
    (cell-ref (cell-lookup var vars vals))))

(define assign!  ;; name changed
  ;; TYPE: (-> (symbol Expressed-Value (list symbol) (vector Denoted-Value))
  ;;           void)
  (lambda (var value vars vals)
    (cell-set! (cell-lookup var vars vals) value)))

(define cell-lookup
  ;; TYPE: (-> (symbol (list symbol) (vector Denoted-Value)) Denoted-Value)
  (lambda (var vars vals)
    (letrec ((loop
	      ;; TYPE:(-> ((list symbol) number)
	      ;;           Denoted-Value)
	      (lambda (vars c)
		(cond
		 ((null? vars) (error "Unassigned variable:" var))
		 ((eq? (car vars) var) (vector-ref vals c))
		 (else (loop (cdr vars) (- c 1)))))))
      (loop vars (- (length vars) 1)))))

(define meth-call
  ;; TYPE: (-> (symbol Class (list Expressed-Value)) Expressed-Value)
  (lambda (name class args)
    (let ((method (meth-lookup name class)))
      (method args))))

(define meth-lookup
  ;; TYPE: (-> (symbol Class) Method)
  (lambda (name class)
    (apply-method-env (class->m-env class) name))) 

(define make-vals
  ;; TYPE: (-> ((list symbol)) (vector Denoted-Value))
  (lambda (vars)
    (list->vector (map (lambda (x) (expressed->denoted (number->expressed 0)))
		       vars))))

(define init-meth-env ; corrected for second printing
  ;; TYPE: Method-Environment
  (extend-method-env 
    '(initialize)
    (list (lambda (args) (void->expressed (displayln "Not initialized"))))
    the-empty-method-env))

(define apply-prim-op
  ;; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (number->expressed (+ (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((*) (number->expressed (* (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((-) (number->expressed (- (expressed->number (car args))
				 (expressed->number (cadr args)))) )
      ((add1) (number->expressed (+ (expressed->number (car args)) 1)) )
      ((sub1) (number->expressed (- (expressed->number (car args)) 1)) )
      ;; primitive operators added to interpreter follow
      ((minus) (number->expressed (- (expressed->number (car args)))))
      ((cons) (list->expressed (cons (car args)
				     (expressed->list (cadr args)))))
      ((car) (car (expressed->list (car args))))
      ((cdr) (list->expressed (cdr (expressed->list (car args)))))
      ((cadr) (cadr (expressed->list (car args))))
      ((caddr) (caddr (expressed->list (car args))))
      ((cadddr) (cadddr (expressed->list (car args))))
      ((list) (list->expressed args))
      ((null) (number->expressed (boolean->int
				  (null? (expressed->list (car args))))) )
      ((eq) (number->expressed (boolean->int (eq? (car args) (cadr args)))) )
      ((equal) (number->expressed (boolean->int
				   (= (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((less) (number->expressed (boolean->int
				   (< (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((greater) (number->expressed (boolean->int
				   (> (expressed->number (car args))
				      (expressed->number (cadr args))))) )
      ((zero) (number->expressed (boolean->int
				  (zero? (expressed->number (car args))))) )
      ((write) (void->expressed (write (car args))))
      ((print) (void->expressed (apply displayln args)))
      ((writeln) (void->expressed (begin (for-each (lambda (arg)
						     (write arg)
						     (display " "))
						   args)
					 (newline))))
      ((space) (void->expressed (display " ")))
      ((newline) (void->expressed (newline)))
      ((printchar) (void->expressed
		    (display (integer->char
			      (expressed->number (car args))))))
      ((instance->class) (instance->class (car args)))
      ((class->parent)
       (maybe-test (class->parent (car args))
		   (lambda (x) x)
		   (lambda () (error "no parent for this class!"))))
      ((error) (error "Defined language program called error"))
      ((noop) (number->expressed 0))
      (else (error "Invalid prim-op name:" prim-op)))))

(define prim-op-names  ; TYPE: (list symbol)
  ;; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr print write space newline
      printchar
      instance->class class->parent error noop))

(define init-env
  ;; TYPE: Environment
  (extend-env
   prim-op-names
   (map (compose expressed->denoted
		 (compose procedure->expressed make-prim-proc))
	prim-op-names)
   ;; modified to add the empty list
   (extend-env
    '(emptylist)
    (list (expressed->denoted (list->expressed '())))
    the-empty-env)))

(define base-env
  ;; TYPE: Environment
  init-env) ; init-env of Chapter 5

;;; Modified from Figure 7.2.4 : p. 288

(set! init-env
  (extend-env
    '(baseobject parentof classof)
    (map expressed->denoted
         (list (make-class (make-nothing) '() '#() '() init-meth-env)
	       (procedure->expressed (make-prim-proc 'class->parent))
               (procedure->expressed (make-prim-proc 'instance->class))))
    base-env))
