;; $Id: ch9-base.scm,v 1.2 1999/08/11 16:28:01 leavens Exp $

;; Common code for EOPL chapter 9 interpreters
;; using the chararacter string parser.

;;; Syntactic sugars
;;       see the file below

(load-from-lib "syntax-expand-letrec.scm")

;;; Parsing

;; To use "run", you also need to do
;;        (load "run.scm")
;; To use "read-eval-print", you also need to do
;;        (load "rep.scm")
;; Note that you can't use both of these at the same time.

(load-from-lib "undefine-read-eval-print.scm")
(load-from-lib "undefine-run.scm")
(load-from-lib "ch9-char-parser.scm")

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = Number + Procedure
;;; Expressed-Value = Number + Procedure
;;;                   + Pair(Expressed-Value Expressed-Value) + Null
;;; Procedure = prim-proc + closure

;;; ADTs used

(load-from-lib "cell.scm")
(load-from-lib "environments.scm")

;;; Anonymous figure : page 142, 143

(define-record prim-proc (prim-op))

;;; Anonymous figure : page 146

(define true-value?
  ; TYPE: (-> (integer) boolean)
  (lambda (x)
    (not (zero? x))))

(define boolean->int
  ; TYPE: (-> (boolean) integer)
  (lambda (x)
    (if  x  1  0)))

;;; Anonymous figure : page 152, 153

(define-record closure (formals body env))

(define apply-proc
  ; TYPE: (-> (Procedure (list Expressed-Value)) Expressed-Value)
  (lambda (proc args)
    (variant-case proc
      (prim-proc (prim-op) (apply-prim-op prim-op args))
      (closure (formals body env)
        (eval-exp body (extend-env formals args env)))
      (else (error "Invalid procedure:" proc)))))

(define apply-prim-op
  ; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (+ (car args) (cadr args)))
      ((-) (- (car args) (cadr args)))
      ((*) (* (car args) (cadr args)))
      ((add1) (+ (car args) 1))
      ((sub1) (- (car args) 1))
      ; primitive operators added to interpreter follow
      ((minus) (- (car args)))
      ((cons) (cons (car args) (cadr args)))
      ((car) (car (car args)))
      ((cdr) (cdr (car args)))
      ((cadr) (cadr (car args)))
      ((caddr) (caddr (car args)))
      ((cadddr) (cadddr (car args)))      
      ((list) args)  
      ((null) (boolean->int (null? (car args))) )
      ((eq) (boolean->int (eq? (car args) (cadr args))) ) 
      ((equal) (boolean->int (= (car args) (cadr args))) )
      ((less) (boolean->int (< (car args) (cadr args))) ) 
      ((greater) (boolean->int (> (car args) (cadr args))) )
      ((zero) (boolean->int (zero? (car args))) )
      ((write) (write (car args)))
      ((print) (apply displayln args))
      ((writeln) (begin (for-each (lambda (arg)
				    (write arg)
				    (display " "))
				  args)
			(newline)))
      ((space) (display " "))
      ((newline) (newline))
      ((printchar) (display (integer->char (car args))))
      (else (error "Invalid prim-op name:" prim-op)))))

(define prim-op-names  ; TYPE: (list symbol)
  ; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr write print writeln space newline
      printchar))

(define init-env       ; TYPE: Environment
  (extend-env
    prim-op-names
    (map make-prim-proc prim-op-names)
    (extend-env
     '(emptylist)
      (list '())
      the-empty-env)))

