;;; $Id: combinator-term.scm,v 1.1 1999/02/11 06:25:19 leavens Exp $

;;; An ADT for combinator terms

;;;  <combinator-term> ::= <var> | S | K | I
;;;                     | (<combinator-term> <combinator-term>)
;;;  <var> ::= <symbol>                  

(define-record var (symbol))
(define-record S ())
(define-record K ())
(define-record I ())
(define-record capp (rator rand))

(define combinator-term?
  ;; TYPE: (-> (datum) boolean)
  (lambda (exp)
    (or (var? exp)
	(S? exp)
	(K? exp)
	(I? exp)
	(and (capp? exp)
	     (combinator-term? (capp->rator exp))
	     (combinator-term? (capp->rand exp))))))

(define parse-combinator-term
  ;; TYPE: (-> (datum) combinator-term)
  (lambda (exp)
    (cond
     ((symbol? exp)
      (case exp
       ((S) (make-S))
       ((K) (make-K))
       ((I) (make-I))
       (else (make-var exp))))
     ((and (list? exp) (= (length exp) 2))
      (make-capp (parse-combinator-term (car exp))
		 (parse-combinator-term (cadr exp))))
     (else (error "parse-combinator-term: bad syntax in expression:"
		  exp)))))

(define unparse-combinator-term
  ;; (-> (combinator-term) datum)
  (lambda (exp)
    (variant-case exp
      (S () 'S)
      (K () 'K)
      (I () 'I)
      (var (symbol) symbol)
      (capp (rator rand) (list (unparse-combinator-term rator)
			       (unparse-combinator-term rand)))
      (else (error "unparse-combinator-term: Invalid abstract syntax"
		   exp)))))



