;;; $Id: lambda-1+quote-examples.scm,v 1.5 1999/01/22 17:36:04 leavens Exp $
;;; Examples in the language lambda-1+quote.
;;;  <lambda-1+quote-exp> ::= <varref>
;;;                          | (quote <symbol>)
;;;                          | (lambda (<var>) <lambda-1+quote-exp>)
;;;                          | (<lambda-1+quote-exp> <lambda-1+quote-exp>)
;;; <varref> ::= <var>
;;; <var> ::= <symbol>

(load-from-lib "lambda-1+quote.scm")

(define count-varrefs
  ;; TYPE: (-> (lambda-1+quote-exp) number)
  (lambda (exp)
    ;; ENSURES: result is the number of <varref>s in exp
    (if (varref? exp)
	1
	(if (quote? exp)
	    0
	    (if (lambda? exp)
		(count-varrefs (lambda->body exp))
		(+ (count-varrefs (app->rator exp))
		   (count-varrefs (app->rand exp))))))))

(define all-varrefs
  ;; TYPE: (-> (lambda-1+quote-exp) (list symbol))
  (lambda (exp)
    ;; ENSURES: result is a list of all the <varref>s in exp
    ;;          (this list may contain duplicate occurrences of symbols).
    (if (varref? exp)
	(list (varref->var exp))
	(if (quote? exp)
	    '()
	    (if (lambda? exp)
		(all-varrefs (lambda->body exp))
		(append (all-varrefs (app->rator exp))
			(all-varrefs (app->rand exp))))))))

;;; Note: in class we did all-varrefs so it returned a set,
;;; so we used set-union instead of append.

(define subst-varrefs
  ;; TYPE: (-> (symbol symbol lambda-1+quote-exp) lambda-1+quote-exp)
  (lambda (new old exp)
    ;; ENSURES: result is the same expression as exp, but with all varrefs
    ;;          that are the same as old replaced by new.
    (if (varref? exp)
	(if (eq? old (varref->var exp))
	    (make-varref new)
	    exp)          ; equal & slower: (make-varref (varref->var exp))
	(if (quote? exp)
	    exp           ; equal & slower: (make-quote (quote->symbol exp))
	    (if (lambda? exp)
		(make-lambda
		 (lambda->formal exp)
		 (subst-varrefs new old (lambda->body exp)))
		(make-app
		 (subst-varrefs new old (app->rator exp))
		 (subst-varrefs new old (app->rand exp))))))))
