;;; $Id: lambda-cond-letrec-exp.scm,v 1.4 1999/01/22 17:36:04 leavens Exp $
;;; An ADT for the untyped lambda multiple-argument calculus
;;; with quoted symbols, if expressions, and let.

;;;	<lambda-cond-letrec-exp> ::= <varref>
;;;               | ( quote <symbol> )
;;;		  | ( lambda ( {<var>}* ) <lambda-cond-letrec-exp> )
;;;		  | ( if <lambda-cond-letrec-exp> <lambda-cond-letrec-exp>
;;;                                          <lambda-cond-letrec-exp> )
;;;		  | ( cond {<cond-clause>}+
;;;                        ( else <lambda-cond-letrec-exp> ) )
;;;               | ( let ( {<decl>}+ ) <lambda-cond-letrec-exp> )
;;;               | ( letrec ( {<decl>}+ ) <lambda-cond-letrec-exp> )
;;;		  | ( <lambda-cond-letrec-exp> {<lambda-cond-letrec-exp>}* )
;;;     <decl> ::= ( <var> <lambda-cond-letrec-exp> )
;;;     <cond-clause> ::= ( <lambda-cond-letrec-exp> <lambda-cond-letrec-exp> )
;;;     <var> ::= <symbol>
;;;     <varref> ::= <symbol>

(load-quietly-from-lib "all.scm")
(load-from-lib "decl.scm")
(load-from-lib "cond-clause")

(define lambda-cond-letrec-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (exp)
    (or (varref? exp)
	(and (quote? exp) 
	     (symbol? (quote->symbol exp)))
	(and (lambda? exp) 
	     (lambda-cond-letrec-exp? (lambda->body exp)))
	(and (if? exp)
	     (lambda-cond-letrec-exp? (if->test-exp exp))
	     (lambda-cond-letrec-exp? (if->then-exp exp))
	     (lambda-cond-letrec-exp? (if->else-exp exp)))
	(and (cond? exp)
	     (all (lambda (clause)
		    (and (cond-clause? clause)
			 (lambda-if-letrec-exp? (cond-clause->test clause))
			 (lambda-if-letrec-exp? (cond-clause->body clause))))
		  (cond->clauses exp))
	     (lambda-cond-letrec-exp? (cond->else-exp exp)))
	(and (or (let? exp) (letrec? exp))
	     (>= (length (let->decls exp)) 1)
	     (all (lambda (decl)
		    (and (decl? decl)
			 (lambda-cond-letrec-exp? (decl->exp decl))))
		  (let->decls exp)))
	(and (app? exp)
	     (lambda-cond-letrec-exp? (app->rator exp))
	     (all lambda-cond-letrec-exp? (app->rands exp))))))

(define varref?
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
 symbol?)

(define quote?  
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 2)
	 (eq? (car exp) 'quote)
	 (symbol? (cadr exp)))))

(define lambda? 
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'lambda)
	 (let ((formals (cadr exp)))
	   (and (list? formals)
		(all symbol? formals))))))

(define if? 
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 4)
	 (eq? (car exp) 'if))))

(define lambda-cond-letrec-exp:last-item
  ;; TYPE: (-> ((list T)) T)
  (lambda (ls)
    ;; REQUIRES: ls is not empty
    (if (null? (cdr ls))
	(car ls)
	(lambda-cond-letrec-exp:last-item (cdr ls)))))

(define cond? 
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (>= (length exp) 3)
	 (eq? (car exp) 'cond)
	 (let ((else-clause (lambda-cond-letrec-exp:last-item exp)))
	   (and (list? else-clause)
		(= (length else-clause) 2)
		(eq? (car else-clause) 'else))))))

(define let? 
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'let))))

(define letrec? 
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'letrec))))

(define app? 
  ;; TYPE: (-> (lambda-cond-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (>= (length exp) 1)
	 (not (memq (car exp) '(quote lambda let letrec if cond))))))

(define varref->var
  ;; TYPE: (-> (lambda-cond-letrec-exp) symbol)
  ;; REQUIRES: the argument is a varref
  (lambda (vr) vr))

(define quote->symbol
  ;; TYPE: (-> (lambda-cond-letrec-exp) symbol)
  ;; REQUIRES: the argument is a quoted expression
  cadr)

(define lambda->formals
  ;; TYPE: (-> (lambda-cond-letrec-exp) (list symbol))
  ;; REQUIRES: the argument is a lambda
  (lambda (exp)
    (cadr exp)))

(define lambda->body
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is a lambda
  (lambda (exp)
    (caddr exp)))

(define app->rator
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is an application
  (lambda (exp)
    (car exp)))

(define app->rands
  ;; TYPE: (-> (lambda-cond-letrec-exp) (list lambda-cond-letrec-exp))
  ;; REQUIRES: the argument is an application
  (lambda (exp)
    (cdr exp)))

(define if->test-exp
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is an if
  cadr)

(define if->then-exp
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is an if
  caddr)

(define if->else-exp
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is an if
  cadddr)

(define lambda-cond-letrec-exp:but-last
  ;; TYPE: (-> ((list T)) (list T))
  (lambda (ls)
    ;; REQUIRES: ls is not empty
    (if (null? (cdr ls))
	'()
	(cons (car ls) (lambda-cond-letrec-exp:but-last (cdr ls))))))

(define cond->clauses
  ;; TYPE: (-> (lambda-cond-letrec-exp) (list (cond-clause lambda-cond-letrec-exp)))
  (lambda (exp)
  ;; REQUIRES: the argument is a cond
    (lambda-cond-letrec-exp:but-last (cdr exp))))

(define cond->else-exp
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  (lambda (exp)
    ;; REQUIRES: the argument is a cond
    (cadr (lambda-cond-letrec-exp:last-item (cdr exp)))))

(define let->decls
  ;; TYPE: (-> (lambda-cond-letrec-exp) (list (decl lambda-cond-letrec-exp)))
  ;; REQUIRES: the argument is a let
  cadr)

(define let->body
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is a let
  caddr)

(define letrec->decls
  ;; TYPE: (-> (lambda-cond-letrec-exp) (list (decl lambda-cond-letrec-exp)))
  ;; REQUIRES: the argument is a letrec
  cadr)

(define letrec->body
  ;; TYPE: (-> (lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  ;; REQUIRES: the argument is a letrec
  caddr)

(define make-varref
  ;; TYPE: (-> (symbol) lambda-cond-letrec-exp)
  (lambda (sym)
    sym))

(define make-quote
  ;; TYPE: (-> (symbol) lambda-cond-letrec-exp)
  (lambda (sym)
    (list 'quote sym)))

(define make-lambda
  ;; TYPE: (-> ((list symbol) lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  (lambda (formals body)
    (list 'lambda formals body)))

(define make-if
  ;; TYPE: (-> (lambda-cond-letrec-exp lambda-cond-letrec-exp lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  (lambda (test consequent alternative)
    (list 'if test consequent alternative)))

(define make-cond
  ;; TYPE: (-> ((list (cond-clause lambda-cond-letrec-exp)) lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  (lambda (clauses alternative)
    (cons 'cond
	  (append clauses (list (list 'else alternative))))))

(define make-let
  ;; TYPE: (-> ((list (decl lambda-cond-letrec-exp)) lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  (lambda (decls body)
    (list 'let decls body)))

(define make-letrec
  ;; TYPE: (-> ((list (decl lambda-cond-letrec-exp)) lambda-cond-letrec-exp) lambda-cond-letrec-exp)
  (lambda (decls body)
    (list 'letrec decls body)))

(define make-app
  ;; TYPE: (-> (lambda-cond-letrec-exp (list lambda-cond-letrec-exp)) lambda-cond-letrec-exp)
  cons)
  
(define parse-lambda-cond-letrec-exp
  ;; TYPE: (-> (datum) lambda-cond-letrec-exp)
  (lambda (exp)
    (let ((parse-decl (parse-decl-maker parse-lambda-cond-letrec-exp))
          (parse-cond-clause (parse-cond-clause-maker
			      parse-lambda-cond-letrec-exp)))
      (cond
       ((symbol? exp) (make-varref exp))
       ((quote? exp) (make-quote (cadr exp)))
       ((lambda? exp)
	(make-lambda (cadr exp) (parse-lambda-cond-letrec-exp (caddr exp))))
       ((if? exp)
	(make-if (parse-lambda-cond-letrec-exp (cadr exp))
		 (parse-lambda-cond-letrec-exp (caddr exp))
		 (parse-lambda-cond-letrec-exp (cadddr exp))))
       ((cond? exp)
	(make-cond
	 (map parse-cond-clause (lambda-cond-letrec-exp:but-last (cdr exp)))
	 (cadr (lambda-cond-letrec-exp:last-item (cdr exp)))))
       ((let? exp)
	(make-let (map parse-decl (cadr exp))
		  (parse-lambda-cond-letrec-exp (caddr exp))))
       ((letrec? exp)
	(make-letrec (map parse-decl (cadr exp))
		     (parse-lambda-cond-letrec-exp (caddr exp))))
       ((app? exp)
	(make-app (parse-lambda-cond-letrec-exp (car exp))
		  (map parse-lambda-cond-letrec-exp (cdr exp))))
       (else
	(error "parse-lambda-cond-letrec-exp: bad syntax in expression: " exp))))))
