;;; $Id: lambda-if-letrec-exp.scm,v 1.3 1999/01/22 17:36:04 leavens Exp $
;;; An ADT for the untyped lambda multiple-argument calculus
;;; with quoted symbols, if expressions, and let.

;;;	<lambda-if-letrec-exp> ::= <varref>
;;;               | ( quote <symbol> )
;;;		  | ( lambda ( {<var>}* ) <lambda-if-letrec-exp> )
;;;		  | ( if <lambda-if-letrec-exp> <lambda-if-letrec-exp>
;;;                                          <lambda-if-letrec-exp> )
;;;               | ( let ( {<decl>}+ ) <lambda-if-letrec-exp> )
;;;               | ( letrec ( {<decl>}+ ) <lambda-if-letrec-exp> )
;;;		  | ( <lambda-if-letrec-exp> {<lambda-if-letrec-exp>}* )
;;;     <decl> ::= ( <var> <lambda-if-letrec-exp> )
;;;     <var> ::= <symbol>
;;;     <varref> ::= <symbol>

(load-quietly-from-lib "all.scm")
(load-from-lib "decl.scm")

(define lambda-if-letrec-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (exp)
    (or (varref? exp)
	(and (quote? exp) 
	     (symbol? (quote->symbol exp)))
	(and (lambda? exp) 
	     (lambda-if-letrec-exp? (lambda->body exp)))
	(and (if? exp)
	     (lambda-if-letrec-exp? (if->test-exp exp))
	     (lambda-if-letrec-exp? (if->then-exp exp))
	     (lambda-if-letrec-exp? (if->else-exp exp)))
	(and (or (let? exp) (letrec? exp))
	     (>= (length (let->decls exp)) 1)
	     (all (lambda (decl)
		    (and (decl? decl)
			 (lambda-if-letrec-exp? (decl->exp decl))))
		  (let->decls exp)))
	(and (app? exp)
	     (lambda-if-letrec-exp? (app->rator exp))
	     (all lambda-if-letrec-exp? (app->rands exp))))))

(define varref?
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
 symbol?)

(define quote?  
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 2)
	 (eq? (car exp) 'quote)
	 (symbol? (cadr exp)))))

(define lambda? 
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'lambda)
	 (let ((formals (cadr exp)))
	   (and (list? formals)
		(all symbol? formals))))))

(define if? 
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 4)
	 (eq? (car exp) 'if))))

(define let? 
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'let))))

(define letrec? 
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (= (length exp) 3)
	 (eq? (car exp) 'letrec))))

(define app? 
  ;; TYPE: (-> (lambda-if-letrec-exp) boolean)
  (lambda (exp)
    (and (list? exp)
	 (>= (length exp) 1)
	 (let ((operator (car exp)))
	   (and (not (eq? operator 'quote))
		(not (eq? operator 'lambda))
		(not (eq? operator 'let))
		(not (eq? operator 'letrec))
		(not (eq? operator 'if)))))))

(define varref->var
  ;; TYPE: (-> (lambda-if-letrec-exp) symbol)
  ;; REQUIRES: the argument is a varref
  (lambda (vr) vr))

(define quote->symbol
  ;; TYPE: (-> (lambda-if-letrec-exp) symbol)
  ;; REQUIRES: the argument is a quoted expression
  cadr)

(define lambda->formals
  ;; TYPE: (-> (lambda-if-letrec-exp) (list symbol))
  ;; REQUIRES: the argument is a lambda
  (lambda (exp)
    (cadr exp)))

(define lambda->body
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is a lambda
  (lambda (exp)
    (caddr exp)))

(define app->rator
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is an application
  (lambda (exp)
    (car exp)))

(define app->rands
  ;; TYPE: (-> (lambda-if-letrec-exp) (list lambda-if-letrec-exp))
  ;; REQUIRES: the argument is an application
  (lambda (exp)
    (cdr exp)))

(define if->test-exp
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is an if
  cadr)

(define if->then-exp
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is an if
  caddr)

(define if->else-exp
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is an if
  cadddr)

(define let->decls
  ;; TYPE: (-> (lambda-if-letrec-exp) (list (decl lambda-if-letrec-exp)))
  ;; REQUIRES: the argument is a let
  cadr)

(define let->body
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is a let
  caddr)

(define letrec->decls
  ;; TYPE: (-> (lambda-if-letrec-exp) (list (decl lambda-if-letrec-exp)))
  ;; REQUIRES: the argument is a let
  cadr)

(define letrec->body
  ;; TYPE: (-> (lambda-if-letrec-exp) lambda-if-letrec-exp)
  ;; REQUIRES: the argument is a let
  caddr)

(define make-varref
  ;; TYPE: (-> (symbol) lambda-if-letrec-exp)
  (lambda (sym)
    sym))

(define make-quote
  ;; TYPE: (-> (symbol) lambda-if-letrec-exp)
  (lambda (sym)
    (list 'quote sym)))

(define make-lambda
  ;; TYPE: (-> ((list symbol) lambda-if-letrec-exp) lambda-if-letrec-exp)
  (lambda (formals body)
    (list 'lambda formals body)))

(define make-if
  ;; TYPE: (-> (lambda-if-letrec-exp lambda-if-letrec-exp lambda-if-letrec-exp) lambda-if-letrec-exp)
  (lambda (test consequent alternative)
    (list 'if test consequent alternative)))

(define make-let
  ;; TYPE: (-> ((list (decl lambda-if-letrec-exp)) lambda-if-letrec-exp) lambda-if-letrec-exp)
  (lambda (decls body)
    (list 'let decls body)))


(define make-letrec
  ;; TYPE: (-> ((list (decl lambda-if-letrec-exp)) lambda-if-letrec-exp) lambda-if-letrec-exp)
  (lambda (decls body)
    (list 'letrec decls body)))

(define make-app
  ;; TYPE: (-> (lambda-if-letrec-exp (list lambda-if-letrec-exp)) lambda-if-letrec-exp)
  cons)
  
(define parse-lambda-if-letrec-exp
  ;; TYPE: (-> (datum) lambda-if-letrec-exp)
  (lambda (exp)
    (let ((parse-decl (parse-decl-maker parse-lambda-if-letrec-exp)))
      (cond
       ((symbol? exp) (make-varref exp))
       ((quote? exp) (make-quote (cadr exp)))
       ((lambda? exp)
	(make-lambda (cadr exp) (parse-lambda-if-letrec-exp (caddr exp))))
       ((if? exp)
	(make-if (parse-lambda-if-letrec-exp (cadr exp))
		 (parse-lambda-if-letrec-exp (caddr exp))
		 (parse-lambda-if-letrec-exp (cadddr exp))))
       ((let? exp)
	(make-let (map parse-decl (cadr exp))
		  (parse-lambda-if-letrec-exp (caddr exp))))
       ((letrec? exp)
	(make-letrec (map parse-decl (cadr exp))
		     (parse-lambda-if-letrec-exp (caddr exp))))
       ((app? exp)
	(make-app (parse-lambda-if-letrec-exp (car exp))
		  (map parse-lambda-if-letrec-exp (cdr exp))))
       (else
	(error "parse-lambda-if-letrec-exp: bad syntax in expression: " exp))))))
