;; $Id: monad-ch5-4.scm,v 1.1 1999/08/11 16:22:55 leavens Exp $

;; Code for EOPL section 5.4 using the chararacter string parser and monads,
;; and the syntactic sugar for let.

;;; Grammar:
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;                  | if <exp> then <exp> else <exp>
;;	            | proc <varlist> <exp>
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( )
;;	            | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*

;;; Syntactic sugars
;;        let x1 = e1; ...; xn = en in body
;;    ==> (proc (x1, ..., xn) body) (e1, ..., en)

(load-from-lib "syntax-expand-let.scm")

;;; Top-level Forms
;;      <form> ::= <exp>

(load-from-lib "monad-eval-form-0.scm")

;;; Domains for this interpreter:
;;;
;;; Denoted-Value = Expressed-Value
;;; Expressed-Value = Number + Procedure
;;;                   + Pair(Expressed-Value Expressed-Value) + Null
;;; Procedure = prim-proc + closure

;; Monad used: Env

(load-from-lib "monad-env.scm")
(define bind bind-env)
(define return return-env)
(define show show-env)

(load-from-lib "monad-base-interpreter.scm")

;;; adapted from Figure 5.4.1 : page 154

(define eval-exp
  ; TYPE: (-> (parsed-exp) (Env Expressed-Value))
  (lambda (exp)
    (variant-case exp
      (lit (datum) (return datum))
      (varref (var)
	  (lookup-env var))
      (app (rator rands)
	  (bind (eval-exp rator)
		(lambda (proc)
		  (bind (eval-rands rands)
			(lambda (args)
			  (apply-proc proc args))))))
      (if (test-exp then-exp else-exp)
	  (bind (eval-exp test-exp)
		(lambda (b)
		  (if (true-value? b)
		      (eval-exp then-exp)
		      (eval-exp else-exp)))))
      (proc (formals body)
	    (close-with-env formals body))
      (else (error "Invalid abstract syntax:" exp)))))

;;; adapted from Anonymous figure : page 152, 153

(define apply-proc
  ; TYPE: (-> (Procedure (list Expressed-Value))
  ;           (Env Expressed-Value))
  (lambda (proc args)
    (variant-case proc
      (prim-proc (prim-op) (return (apply-prim-op prim-op args)))
      (closure (formals body env)
	(scope-env formals args env
	    (eval-exp body)))
      (else (error "Invalid procedure:" proc)))))
