;; $Id: monad-ch5-5.scm,v 1.1 1999/08/11 16:22:56 leavens Exp $

;; Code for EOPL section 5.5 using the chararacter string parser and monads,
;; and the syntactic sugar for let.

;;; Grammar:
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;                  | if <exp> then <exp> else <exp>
;;	            | proc <varlist> <exp>
;;                  | <var> := <exp>
;;                  | begin <exp> {; <exp>}* end
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( )
;;	            | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*

;;; Syntactic sugars
;;        let x1 = e1; ...; xn = en in body
;;    ==> (proc (x1, ..., xn) body) (e1, ..., en)

(load-from-lib "syntax-expand-let.scm")

;;; Top-level Forms
;;      <form> ::= <exp>

(load-from-lib "monad-eval-form-0.scm")

;;; Domains for this interpreter:
;;;
;;; Denoted-Value = Cell(Expressed-Value)
;;; Expressed-Value = Number + Procedure
;;;                   + Pair(Expressed-Value Expressed-Value) + Null
;;; Procedure = prim-proc + closure

;; Monad used: Env o Store

(load-from-lib "monad-env+store.scm")
(define return return-env+store)
(define bind bind-env+store)
(define show show-env+store)

(load-from-lib "monad-base-interpreter.scm")

;;; adapted from Figure 5.5.1: page 157

(define eval-exp
  ; TYPE: (-> (parsed-exp) (Env (Store Expressed-Value)))
  (lambda (exp)
    (variant-case exp
      (lit (datum) (return datum))
      (varref (var)
	  (var-value var))
      (app (rator rands)
	  (bind (eval-exp rator)
		(lambda (proc)
		  (bind (eval-rands rands)
			(lambda (args)
			  (apply-proc proc args))))))
      (if (test-exp then-exp else-exp)
	  (bind (eval-exp test-exp)
		(lambda (b)
		  (if (true-value? b)
		      (eval-exp then-exp)
		      (eval-exp else-exp)))))
      (proc (formals body)
	    (close-with-env formals body))
      (varassign (var exp)
	  (bind (var-location var)
		(lambda (loc)
		  (bind (eval-exp exp)
			(lambda (val)
			  (update-store (list loc) (list val)))))))
      (begin (exp1 exp2)
	(bind (eval-exp exp1)
	      (lambda (ignored)
		(eval-exp exp2))))
      (else (error "Invalid abstract syntax:" exp)))))

;;; adapted from Figure 5.5.1: page 157

;;; Procedure = prim-proc + closure

(define apply-proc
  ; TYPE: (-> (Procedure (list Expressed-Value))
  ;           (Env (Store Expressed-Value)))
  (lambda (proc args)
    (variant-case proc
      (prim-proc (prim-op) (return (apply-prim-op prim-op args)))
      (closure (formals body env)
	(scope-env formals args env
	    (eval-exp body)))
      (else (error "Invalid procedure:" proc)))))
