;;; $Id: monad-env+store.scm,v 1.1 1999/08/11 16:22:57 leavens Exp $
;;; The Env+Store monad
;;;     (Env+Store T) = (Env (Store T))
;;;                   = (-> (Environment) (-> (interp-result T State)))
;;; see Wadler's paper "The Essence of Functional Programming".

;;; Domains:
;;;
;;; Environment = <var> -> Denoted-Value

;;; ADTs used

(load-from-lib "environments.scm")
(load-from-lib "stores.scm")

;;; Anonymous figure : page 152, 153

(define-record closure (formals body env))

;;; Anonymous figure : page 142, 143

(define-record prim-proc (prim-op))

;;; from EOPL, page 162
(define-record interp-result (value store))

(define return-env+store
  ; TYPE: (-> (T) (Env (Store T)))
  (lambda (a)
    (lambda (env)
      (lambda (s0)
	(make-interp-result a s0)))))

(define bind-env+store
  ; TYPE: (-> ((Env (Store T)) (-> (T) (Env (Store U)))) (Env (Store U)))
  (lambda (m k)
    (lambda (env)
      (lambda (s0)
	(let* ((res ((m env) s0))
	       (a (interp-result->value res))
	       (s1 (interp-result->store res)))
	  ; uncommenting the following gives a trace useful for debugging
	  ; (displayln "bind-env+store: k = " k)
	  ; (displayln "bind-env+store: a = " a)
	  ; ;(displayln "bind-env+store: s1 = " s1)
	  (((k a) env) s1))))))

(define show-env+store
  ; TYPE: (-> ((Env (Store T))) T)
  (lambda (m)
    (((add-to-env prim-value-names prim-values
	 (add-to-env prim-op-names (map make-prim-proc prim-op-names)
	     (bind-env+store m
		(lambda (res)
		  (lambda (ignored-init-env)
		     (lambda (ignored-final-store)
			res))))))
      the-empty-env)
     the-empty-store)))

(define var-value
  ; TYPE: (-> (symbol) (Env (Store Storable-Value)))
  (lambda (var)
    (lambda (env)
      (lambda (s)
	; (displayln "var-value: var = " var)
	; (displayln "var-value: env = " env)
	(let ((res (apply-store s (apply-env env var))))
	  ; (displayln "var-value: res = " res)
	  (make-interp-result res s))))))

(define var-location
  ; TYPE: (-> (symbol) (Env (Store Location)))
  (lambda (var)
    (lambda (env)
      (lambda (s)
	; (displayln "var-location: var = " var)
	; (displayln "var-location: env = " env)
	(let ((res (apply-env env var)))
	  ; (displayln "var-location: res = " res)
	  (make-interp-result res s))))))

(define allocate-store
  ; TYPE: (-> ((list Storable-Value)) (Env (Store (list Location))))
  (lambda (vals)
    (lambda (env)
      (lambda (s)
	(let ((locs-and-state (extend-store vals s)))
	  (make-interp-result
	   (car locs-and-state)
	   (cdr locs-and-state)))))))

(define update-store
  ; TYPE: (-> ((list Location) (list Storable-Value)) (Env (Store Void)))
  (lambda (locs vals)
    (lambda (env)
      (lambda (s)
	(make-interp-result '() (assign-store locs vals s))))))

(define close-with-env
  ; TYPE: (-> ((list symbol) parsed-exp) (Env (Store Denotable-Value)))
  (lambda (formals body)
    (lambda (env)
      (lambda (store)
	(make-interp-result (make-closure formals body env) store)))))

(define read-env
  ; TYPE: (Env (Store Environment))
  (lambda (env)
    (lambda (s)
      (make-interp-result env s))))

(define in-env
  ; TYPE: (-> (Environment (Env (Store T))) T)
  (lambda (new-env m)
    (lambda (surrounding-env)
       ; (displayln "in-env: new-env = " new-env)
       (m new-env))))

(define scope-env
  ; TYPE: (-> ((list symbol) (list Expressible-Value) Environment
  ;            (Env (Store T)))
  ;           (Env (Store T)))
  (lambda (vars vals old-env m)
    (bind-env+store (allocate-store vals)
       (lambda (val-locs)
	 (in-env (extend-env vars val-locs old-env) m)))))

(define add-to-env
  ; TYPE: (-> ((list symbol) (list Expressible-Value)
  ;            (Env (Store T)))
  ;           (Env (Store T)))
  (lambda (vars vals m)
    (bind-env+store read-env
       (lambda (env)
	 ; (displayln "in add-to-env, env = " env)
	 (bind-env+store (allocate-store vals)
	     (lambda (val-locs)
	       (in-env (extend-env vars val-locs env) m)))))))

(define prim-value-names   ; TYPE: (list symbol)
  '(emptylist))

(define prim-values        ; TYPE: (list Expressed-Value)
  '(()))

;;; Anonymous figure : page 152, 153

(define prim-op-names  ; TYPE: (list symbol)
  ; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr write print writeln space newline
      printchar))

(define apply-prim-op
  ; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (+ (car args) (cadr args)))
      ((-) (- (car args) (cadr args)))
      ((*) (* (car args) (cadr args)))
      ((add1) (+ (car args) 1))
      ((sub1) (- (car args) 1))
      ; primitive operators added to interpreter follow
      ((minus) (- (car args)))
      ((cons) (cons (car args) (cadr args)))
      ((car) (car (car args)))
      ((cdr) (cdr (car args)))
      ((cadr) (cadr (car args)))
      ((caddr) (caddr (car args)))
      ((cadddr) (cadddr (car args)))      
      ((list) args)  
      ((null) (boolean->int (null? (car args))) )
      ((eq) (boolean->int (eq? (car args) (cadr args))) ) 
      ((equal) (boolean->int (= (car args) (cadr args))) )
      ((less) (boolean->int (< (car args) (cadr args))) ) 
      ((greater) (boolean->int (> (car args) (cadr args))) )
      ((zero) (boolean->int (zero? (car args))) )
      ((write) (write (car args)))
      ((print) (apply displayln args))
      ((writeln) (begin (for-each (lambda (arg)
				    (write arg)
				    (display " "))
				  args)
			(newline)))
      ((space) (display " "))
      ((newline) (newline))
      ((printchar) (display (integer->char (car args))))
      (else (error "Invalid prim-op name:" prim-op)))))
