;;; $Id: monad-env.scm,v 1.2 1998/01/26 21:32:29 leavens Exp $
;;; The environment monad,
;;;     (Env T) = (-> (Environment) T)
;;; see Wadler's paper "The Essence of Functional Programming".

;;; Domains:
;;;
;;; Environment = <var> -> Denoted-Value

;;; ADTs used

(load-from-lib "environments.scm")

;;; Anonymous figure : page 152, 153

(define-record closure (formals body env))

;;; Anonymous figure : page 142, 143

(define-record prim-proc (prim-op))


(define return-env
  ; TYPE: (-> (T) (Env T))
  (lambda (a)
    (lambda (env)
      a)))

(define bind-env
  ; TYPE: (-> ((Env T) (-> (T) (Env U))) (Env U))
  (lambda (m k)
    (lambda (env)
      (let ((res (m env)))
	; uncommenting the following gives a trace useful for debugging
	; (displayln "bind-env: k = " k)
	; (displayln "bind-env: res = " res)
	((k res) env)))))

(define show-env
  ; TYPE: (-> ((Env T)) T)
  (lambda (m)
    (let ((init-env ; TYPE: Environment
	   (extend-env
	     prim-op-names
	     (map make-prim-proc prim-op-names)
	     (extend-env
	       prim-value-names
	       prim-values
	       the-empty-env))))
      (m init-env))))

(define lookup-env
  ; TYPE: (-> (var) (Env Denotable-Value))
  (lambda (var)
    (lambda (env)
      (apply-env env var))))

(define close-with-env
  ; TYPE: (-> ((list symbol) parsed-exp) (Env Denotable-Value))
  (lambda (formals body)
    (lambda (env)
      (make-closure formals body env))))

(define read-env
  ; TYPE: (Env Environment)
  (lambda (env)
    (lambda (s)
      env)))

(define in-env
  ; TYPE: (-> (Environment (Env T)) (Env T))
  (lambda (new-env m)
    (lambda (surrounding-env)
       ; (displayln "in-env: new-env = " new-env)
       (m new-env))))

(define scope-env
  ; TYPE: (-> ((list symbol) (list Expressible-Value) Environment
  ;            (Env T)
  ;           (Env T))
  (lambda (vars vals old-env m)
    (in-env (extend-env vars vals old-env) m)))

(define add-to-env
  ; TYPE: (-> ((list symbol) (list Expressible-Value)
  ;            (Env T))
  ;           (Env T))
  (lambda (vars vals m)
    (bind-env read-env
       (lambda (env)
	 ; (displayln "add-to-env: env = " env)
	 (in-env (extend-env vars vals env) m)))))

(define prim-value-names   ; TYPE: (list symbol)
  '(emptylist))

(define prim-values        ; TYPE: (list Expressed-Value)
  '(()))

;;; Anonymous figure : page 152, 153

(define prim-op-names  ; TYPE: (list symbol)
  ; modified to add primitive operators
  '(+ - * add1 sub1 minus cons car cdr list null eq equal less greater zero
      cadr caddr cadddr write print writeln space newline
      printchar))

(define apply-prim-op
  ; TYPE: (-> (symbol (list Expressed-Value)) Expressed-Value)
  (lambda (prim-op args)
    (case prim-op
      ((+) (+ (car args) (cadr args)))
      ((-) (- (car args) (cadr args)))
      ((*) (* (car args) (cadr args)))
      ((add1) (+ (car args) 1))
      ((sub1) (- (car args) 1))
      ; primitive operators added to interpreter follow
      ((minus) (- (car args)))
      ((cons) (cons (car args) (cadr args)))
      ((car) (car (car args)))
      ((cdr) (cdr (car args)))
      ((cadr) (cadr (car args)))
      ((caddr) (caddr (car args)))
      ((cadddr) (cadddr (car args)))      
      ((list) args)  
      ((null) (boolean->int (null? (car args))) )
      ((eq) (boolean->int (eq? (car args) (cadr args))) ) 
      ((equal) (boolean->int (= (car args) (cadr args))) )
      ((less) (boolean->int (< (car args) (cadr args))) ) 
      ((greater) (boolean->int (> (car args) (cadr args))) )
      ((zero) (boolean->int (zero? (car args))) )
      ((write) (write (car args)))
      ((print) (apply displayln args))
      ((writeln) (begin (for-each (lambda (arg)
				    (write arg)
				    (display " "))
				  args)
			(newline)))
      ((space) (display " "))
      ((newline) (newline))
      ((printchar) (display (integer->char (car args))))
      (else (error "Invalid prim-op name:" prim-op)))))
