;;; $Id: monad-store.scm,v 1.1 1999/08/11 16:22:59 leavens Exp $
;;; The store monad,
;;;     (Store T) = (-> (State) (interp-result T State))
;;; adapted from
;;; Wadler's paper "The Essence of Functional Programming", section 2.5

(load-from-lib "stores.scm")

;;; from EOPL, page 162
(define-record interp-result (value store))

(define return-store
  ; TYPE: (-> (T) (Store T))
  (lambda (a)
    (lambda (s0)
      (make-interp-result a s0))))

(define bind-store
  ; TYPE: (-> ((Store T) (-> (T) (Store U))) (Store U))
  (lambda (m k)
    (lambda (s0)
      (let* ((res (m s0))
	     (a (interp-result->value res))
	     (s1 (interp-result->store res)))
	; uncommenting the following gives a trace useful for debugging
	; (displayln "k = " k)
	; (displayln "a = " a)
	; (displayln "s1 = " s1)
	((k a) s1)))))

(define lookup-store
  ; TYPE: (-> (Location) (Store Storable-Value))
  (lambda (loc)
    (lambda (s)
      (make-interp-result (apply-store s loc) s))))

(define allocate-store
  ; TYPE: (-> ((list Storable-Value)) (Store (list Location)))
  (lambda (vals)
    (lambda (s)
      (let ((locs-and-state (extend-store vals s)))
	(make-interp-result
	 (car locs-and-state)
	 (cdr locs-and-state))))))

(define update-store
  ; TYPE: (-> ((list Location) (list Storable-Value)) (Store Void))
  (lambda (locs vals)
    (lambda (s)
      (make-interp-result '() (assign-store locs vals s)))))

(define show-store
  ; TYPE: (-> ((Store T)) T)
  (lambda (m)
      (let* ((res (m the-empty-store))
	     (a (interp-result->value res))
	     (s1 (interp-result->store res)))
	a)))
