;;; $Id: record-lambda-1+number-exp-examples.scm,v 1.1 1998/10/07 23:04:36 leavens Exp $

;;; Grammar:
;;;  <lambda-1+number-exp> ::= <varref>
;;;                          | <number>
;;;                          | (lambda (<var>) <lambda-1+number-exp>)
;;;                          | (<lambda-1+number-exp> <lambda-1+number-exp>)
;;; <varref> ::= <var>
;;; <var> ::= <symbol>

(load-from-lib "record-lambda-1+number-exp.scm")

(define count-varrefs
  ;; TYPE: (-> (lambda-1+number-exp-exp) number)
  (lambda (exp)
    ;; ENSURES: result is the number of <varref>s in exp
    (variant-case exp
      (varref () 1)
      (lit () 0)
      (lambda (body) (count-varrefs body))
      (app (rator rand) (+ (count-varrefs rator)
			   (count-varrefs rand)))
      (else (error "count-varrefs should not happen")))))

(define all-varrefs
  ;; TYPE: (-> (lambda-1+number-exp-exp) (list symbol))
  (lambda (exp)
    ;; ENSURES: result is a list of all the <varref>s in exp
    ;;          (this list may contain duplicate occurrences of symbols).
    (variant-case exp
      (lit (datum) '())
      (varref (var) (list var))
      (lambda (formal body) (all-varrefs body))
      (app (rator rand) (append (all-varrefs rator)
				(all-varrefs rand)))
      (else (error "all-varrefs should not happen")))))

;;; Note: in class we did all-varrefs so it returned a set,
;;; so we used set-union instead of append.


(define subst-varrefs
  ;; TYPE: (-> (symbol symbol lambda-1+number-exp-exp) lambda-1+number-exp-exp)
  (lambda (new old exp)
    ;; ENSURES: result is the same expression as exp, but with all varrefs
    ;;          that are the same as old replaced by new.
    (variant-case exp
      (lit (datum) exp) ; equal & slower: (make-lit datum)
      (varref (var) (if (eq? old var)
			(make-varref new)
			exp)) ; equal & slower: (make-varref var)
      (lambda (formal body)
       (make-lambda formal (subst-varrefs new old body)))
      (app (rator rand)
       (make-app
	(subst-varrefs new old rator)
	(subst-varrefs new old rand)))
      (else (error "subst-varrefs: should not happen")))))

