;;; $Id: record-lambda-1+number-exp.scm,v 1.2 1998/10/06 21:01:24 leavens Exp $
;;; An ADT for the untyped lambda 1-argument calculus with numeric literals

;;; Grammar:
;;;  <lambda-1+number-exp> ::= <varref>
;;;                          | <number>
;;;                          | (lambda (<var>) <lambda-1+number-exp>)
;;;                          | (<lambda-1+number-exp> <lambda-1+number-exp>)
;;; <varref> ::= <var>
;;; <var> ::= <symbol>

(define-record lit (datum))
(define-record varref (var))
(define-record lambda (formal body))
(define-record app (rator rand))

(define lambda-1+number-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (exp)
    (or (lit? exp)
	(varref? exp)
	(and (lambda? exp)
	     (lambda-1+number-exp? (lambda->body exp)))
	(and (app? exp)
	     (lambda-1+number-exp? (app->rator exp))
	     (lambda-1+number-exp? (app->rand exp))))))

(define parse-lambda-1+number-exp
  ;; TYPE: (-> (datum) lambda-1+number-exp)
  (lambda (exp)
    (cond
     ((number? exp) (make-lit exp))
     ((symbol? exp) (make-varref exp))
     ((and (list? exp) (= (length exp) 3) (eq? (car exp) 'lambda))
      (make-lambda (caadr exp)
		   (parse-lambda-1+number-exp (caddr exp))))
     ((and (list? exp) (= (length exp) 2))
      (make-app (parse-lambda-1+number-exp (car exp))
		 (parse-lambda-1+number-exp (cadr exp))))
     (else (error "parse-lambda-1+number-exp: bad syntax in expression:"
		  exp)))))

(define unparse-lambda-1+number-exp
  ;; (-> (lambda-1+number-exp) datum)
  (lambda (exp)
    (variant-case exp
      (lit (datum) datum)
      (varref (var) var)
      (lambda (formal body)
	(list 'lambda (list formal) (unparse-lambda-1+number-exp body)))
      (app (rator rand) (list (unparse-lambda-1+number-exp rator)
			      (unparse-lambda-1+number-exp rand)))
      (else (error "unparse-lambda-1+number-exp: Invalid abstract syntax"
		   exp)))))
