;;; $Id: simple-type-expr.scm,v 1.2 1999/02/08 19:23:49 leavens Exp $
;;;
;;; External types, reported to and gotten from the user:
;;;
;;;   <type> ::= <basic-type> | <type-variable> | <function-type>
;;;   <basic-type> ::= number | boolean | string | character | symbol
;;;   <type-variable> ::= S | T | U | V | W | X | Y | Z
;;;   <function-type> ::= (-> ({<type>}*) <type>)
;;;
;;; (This is a subset of the types used in the course's type checker.
;;; See the file type-check-external-type-expr.scm for the full set.)

(load-quietly-from-lib "all.scm")

(define type-expr?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (or (type-variable? d)
	(basic-type? d)
	(and (function-type? d)
	     (all type-expr? (cadr d))
	     (type-expr? (caddr d)))
	)))

(define basic-type?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (and (symbol? d)
	 (not (not   ; this makes sure a proper boolean is returned
	       (memq d '(number boolean string character symbol)))))))

(define type-variable?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (and (symbol? d)
	 (not (not   ; this makes sure a proper boolean is returned
	       (memq d '(S T U V W X Y Z)))))))

(define function-type?
  ;; TYPE: (-> (type-expr) boolean)
  (lambda (d)
    (and (list? d)
	 (= (length d) 3)
	 (eq? (car d) '->)
	 (list? (cadr d)))))

(define make-basic-type
  ;; TYPE: (-> (symbol) type-expr)
  (lambda (s) s))

(define make-type-variable
  ;; TYPE: (-> (symbol) type-expr)
  (lambda (v) v))

(define make-function-type
  ;; TYPE: (-> ((list type-expr) type-expr) type-expr)
  (lambda (arg-types result-type)
    (list '-> arg-types result-type)))

(define type-variable->symbol
  ;; TYPE: (-> (type-expr) symbol)
  (lambda (t)
    (if (type-variable? t)
	t
	(error "type-variable->symbol called with non-type-variable argument "
	       t))))

(define basic-type->symbol
  ;; TYPE: (-> (type-expr) symbol)
  (lambda (t)
    (if (basic-type? t)
	t
	(error "basic-type->symbol called with non-basic-type argument "
	       t))))

(define function-type->arg-types
  ;; TYPE: (-> (type-expr) (list type-expr))
  (lambda (t)
    (if (function-type? t)
	(cadr t)
	(error
	 "function-type->arg-types called with non-function type argument "
	 t))))

(define function-type->result-type
  ;; TYPE: (-> (type-expr) type-expr)
  (lambda (t)
    (if (function-type? t)
	(caddr t)
	(error
	 "function-type->result-type called with non-function type argument "
	 t))))

(define parse-type-expr
  ;; TYPE: (-> (datum) type-expr)
  (lambda (d)
    (cond
    ((type-variable? d)
     (make-type-variable d))
    ((basic-type? d)
     (make-basic-type d))
    ((function-type? d)
     (make-function-type
      (map parse-type-expr (cadr d))
      (parse-type-expr (caddr d))))
    (else (error "parse-type-expr syntax error in " d)))))
