;;; $Id: sym-exp.scm,v 1.1 1999/01/22 07:46:06 leavens Exp $
;;; "Sym-exps", a form of symbol-expressions,, using the grammar
;;;    <sym-exp> ::= <symbol> | ( {<sym-exp>}* )

(load-quietly-from-lib "all.scm")

(define s-list?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (and (list? d)
	 (all sym-exp? d))))

(define sym-exp?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (or (symbol? d)
	(s-list? d))))

(define symbol->sym-exp
  ;; TYPE: (-> (symbol) sym-exp)
  (lambda (s)
    s))

(define s-list->sym-exp
  ;; TYPE: (-> ((list sym-exp)) sym-exp)
  (lambda (slst)
    slst))

(define sym-exp-symbol?
  ;; TYPE: (-> (sym-exp) boolean)
  symbol?)

(define sym-exp-s-list?
  ;; TYPE: (-> (sym-exp) boolean)
  list?)

(define sym-exp->symbol
  ;; TYPE: (-> (sym-exp) symbol)
  (lambda (se)
    (if (symbol? se)
	se
	(error "sym-exp->symbol: this is not a symbol: " se))))

(define sym-exp->s-list
  ;; TYPE: (-> (sym-exp) (list sym-exp))
  (lambda (se)
    (if (not (symbol? se))
	se
	(error "sym-exp->symbol: this is not a s-list: " se))))

(define parse-sym-exp
  ;; TYPE: (-> (datum) sym-exp)
  (lambda (exp)
    (cond
     ((symbol? exp) (symbol->sym-exp exp))
     ((list? exp) (s-list->sym-exp (parse-s-list exp)))
     (else (error "parse-sym-exp: bad syntax: " exp)))))

(define parse-s-list
  ;; TYPE: (-> (datum) (list sym-exp))
  (lambda (exp)
    (cond
     ((list? exp) (map parse-sym-exp exp))
     (else (error "parse-s-list: bad syntax: " exp)))))


