;;; $Id: type-check-and-eval.scm,v 1.27 1999/06/23 16:57:22 leavens Exp $
;;;---------------------------------------------------------------------
;;; Here is the complete load tree for the type checker.  This load tree
;;; diagram was last updated 5/26/99:
;;;
;;; type-check-and-eval.scm
;;;  |
;;;  |--type-check.scm
;;;      |
;;;      |--type-check-polytype.scm
;;;      |   |
;;;      |   |--type-check-unify.scm
;;;      |   |   |
;;;      |   |   |--maybe.scm
;;;      |   |   |
;;;      |   |   |--type-check-env.scm
;;;      |   |   |
;;;      |   |   |--type-check-type-expr.scm
;;;      |   |       |
;;;      |   |       |--type-check-utils.scm
;;;      |   |           |
;;;      |   |           |--displayln.scm 
;;;      |   |           |
;;;      |   |           |--atomic-tree.scm
;;;      |   |               |
;;;      |   |               |--atomic-expression.scm
;;;      |   |                   |
;;;      |   |                   |--all.scm
;;;      |   |                   |
;;;      |   |                   |--atomic.scm
;;;      |   |
;;;      |   |--type-check-syntax-helpers.scm
;;;      |   |
;;;      |   |--type-check-global-var-types.scm
;;;      |   |
;;;      |   |--type-check-type-translate.scm
;;;      |       |
;;;      |       |--type-check-external-type-expr.scm
;;;      |           |
;;;      |           |--type-check-variable.scm
;;;      |
;;;      |--type-check-down-type.scm
;;;
;;;---------------------------------------------------------------------

;;; Gary: The following definitions allow the insertion of deftypes and
;;; defreps in .scm files.  This isn't enough however as the type-expr
;;; following the varref or rep-type is treated as a s-expression.
(define deftype
  ;; TYPE: (-> (datum ...) void)
  (lambda x (tc:ignore x)))

(define defrep
  ;; TYPE: (-> (datum ...) void)
  (lambda x (tc:ignore x)))

(Load-quietly-from-lib "type-check.scm")

;;; Set the following to #f to suppress evaluation when there is a type error.
(define *eval-even-if-error* #t)

;;; Set the following to #t to enable additional prompting, e.g., prompt
;;; before evaluating type-erroneous expression.  Three procedures defined
;;; at the end of this file provide a public interface to this variable.
(define *tc:prompting-on* #t)

;;; Set the following > 0 to enable verbose tracing of the type-checking
;;; process.  The greater the value the more tracing information displayed.
;;; Three procedures defined at the end of this file provide a
;;; public interface to this variable.
(define *tc:tracing-level* 0)

(define *tc:the-last-name-defined* 'nothing)

(define tc:type-check-to-data
  ;; TYPE: (-> (expression) external-type-expr)
  (tc:type-check-maker
   (lambda (name)
     (set! *tc:the-last-name-defined* name))
   (lambda (typ) typ)))

(define tc:orig-error error)

(define *tc:type-check-and-eval-start*
  ;; TYPE: (-> (datum) poof)
  (lambda (ignored)
    (tc:ignore ignored)
    (displayln "exiting")
    (exit)))

;;; change what error does
(set! error
      ;; TYPE: (-> (datum ...) poof)
      (lambda args
	(let ((cep
	       (current-error-port))); if needed, change to (current-output-port)
	  (display "ERROR: " cep)
	  (if (not (null? args))
	      (begin (display (car args) cep)
		     (for-each (lambda (x)
				 (display " " cep)
				 (write x cep))
			       (cdr args))
		     (newline cep)))
	  (*tc:type-check-and-eval-start* 'ignored))))

(define type-check-and-eval-loop 
  ;; TYPE: (-> () poof)
  (lambda ()
    (call-with-current-continuation
     (lambda (start-continuation)
       ;; calling *tc:type-check-and-eval-start* will go from below next line
       (set! *tc:type-check-and-eval-start* start-continuation)))
    (tc:type-check-and-eval-loop-helper "typed> ")))

(define tc:type-check-and-eval-loop-helper 
  ;; TYPE: (-> (string) poof)
  (lambda (prompt-str)
    ;; EFFECT: prompt with prompt-str, read expressions and tc:type check,
    ;;         until done or EOF is entered
    (display prompt-str)
    (force-output)
    (let ((expr (read)))
      (cond
       ((or (eof-object? expr) (equal? expr '(exit)) (equal? expr '(quit)))
	(display "Bye.")
	(newline)
	(exit))
       ((equal? expr '(type-check-exit))
	(newline)
	(displayln "Use `(type-check-and-eval-loop)' to restart the type checker.")
	(abort))			; This may be SCM dependent.
       (else
	(let* ((typ (tc:type-check-to-data expr))
	       (type-expr (tc:type-untranslate typ))
	       (err-typ? (tc:error-type? type-expr)))
	  ;; (displayln "tc:type-check-and-eval-loop-helper: typ:")
	  ;; (pretty-print typ)
	  (if (and err-typ? (not *eval-even-if-error*))
	      (tc:type-check-and-eval-loop-helper prompt-str)
	      (letrec
		  ((eval-it
		    ;; TYPE: (-> () void)
		    (lambda ()
		      ;;EFFECT: Evaluate expr and print its result (or the name of the object
		      ;;        defined) and type.
		      (cond
		       ((tc:define-exp? expr)
			(begin
			  (eval expr)
			  (display *tc:the-last-name-defined*)))
		       ((tc:grouped-deftypes? expr)
			(display "#<unspecified>"))
		       ((tc:deftype? expr)
			(display *tc:the-last-name-defined*))
		       ((and (not err-typ?)
			     (pair? expr)
			     (or (eq? (car expr) 'load)
				 (eq? (car expr) 'load-quietly)))
			(begin
			  (tc:type-check-file (cadr expr)
					      (eq? (car expr) 'load-quietly))
			  (write (eval expr))))
		       ((and (not err-typ?)
			     (pair? expr)
			     (or (eq? (car expr) 'load-from-lib)
				 (eq? (car expr) 'load-quietly-from-lib)))
			(let ((fullname (string-append *cs342:libdir* (cadr expr))))
			  (tc:type-check-file
			   fullname
			   (eq? (car expr) 'load-quietly-from-lib))
			  (write (eval expr))))
		       (else (write (eval expr))))
		      (if (not err-typ?)
			  (begin
			    (display tc:colon)
			    (display typ)))
		      (newline))))
		(if err-typ?
		    (begin
		      (display "WARNING: this may have a type error...")
		      (if *tc:prompting-on*
			  (begin
			    (newline)
			    (display "Continue despite potential error? (y/n): ")
			    (force-output)
			    (let ((response (read)))
			      (newline)
			      (if (or (equal? response 'y)
				      (equal? response 'yes))
				  (eval-it)
				  (displayln "Evaluation skipped."))))
			  (begin
			    (display "but, since prompting is off, here we go...")
			    (newline)
			    (eval-it))))
		    (eval-it))
		(tc:type-check-and-eval-loop-helper prompt-str)))))))))

;;; The following turns on prompting in the type checker
(define type-prompting-on!
  ;; TYPE: (-> () void)
  (lambda ()
    (set! *tc:prompting-on* #t)
    (displayln "Typed interpreter verbose prompting is now on.")))

;;; The following turns off prompting in the type checker
(define type-prompting-off!
  ;; TYPE: (-> () void)
  (lambda ()
    (set! *tc:prompting-on* #f)
    (displayln "Typed interpreter verbose prompting is now off.")))

;;; The following queries the status of the prompting switch in the
;;; type checker.
(define type-prompting-on?
  ;; TYPE: (-> () boolean)
  (lambda ()
    *tc:prompting-on*))

;;; The following set the level of tracing in the type checker
(define type-tracing-set!
  ;; TYPE: (-> (number) void)
  (lambda (num)
    (set! *tc:tracing-level* num)
    (if (> *tc:tracing-level* 0)
	(displayln "Typed interpreter verbose tracing is now on at level "
		   *tc:tracing-level*)
	(displayln "Typed interpreter verbose tracing is now off."))))

;;; The following turns off tracing in the type checker
(define type-tracing-off!
  ;; TYPE: (-> () void)
  (lambda ()
    (set! *tc:tracing-level* 0)
    (displayln "Typed interpreter verbose tracing is now off.")))

;;; The following queries the status of the tracing level in the
;;; type checker.
(define type-tracing-level
  ;; TYPE: (-> () number)
  (lambda ()
    *tc:tracing-level*))

;;; The following provides a list the public type checker procedures
(define type-help
  ;; TYPE: (-> () void)
  (lambda ()
    (displayln "The Polymorphic Type Checker for Scheme provides ")
    (displayln "the following public procedures:")
    (displayln "type-help:                displays this information")
    (displayln "type-check-and-eval-loop: a typed Scheme interpreter loop")
    (displayln "type-check-exit:          leave the typed interpreter loop")
    (displayln "type-prompting-on!:       enable verbose prompting on errors")
    (displayln "type-prompting-off!:      display verbose prompting on errors")
    (displayln "type-prompting-on?:       returns true if verbose prompting is on")
    (displayln "type-tracing-set!:        set tracing detail level to given value")
    (displayln "type-tracing-off!:        set tracing detail level to 0")
    (displayln "type-tracing-level:       returns the numeric level of tracing detail")
    ))

(displayln
 "Typed Scheme interpreter loop (by Curtis Clifton and Gary T. Leavens)")
(newline)

;;; The following is specific to SCM, remove it if you can't use it
;;; The first 3 elements of *argv* are the program name, and the names
;;; of the two files we use in 342 for setting up the interpreter.
;;; We don't want to load these again, so we take the cdddr of *argv*.
(let ((argv (case (scheme-implementation-type)
	      ((scm) *argv*)
	      ((bigloo) (command-line))
	      (else #f))))
  (if (and (not (equal? argv #f))
	   (>= (length argv) 3))
      (for-each (lambda (fn)
		  (begin
		    (tc:type-check-file fn #f)
		    (load fn)))
		(cdddr argv))))

(type-check-and-eval-loop)
