;;; $Id: type-check-compute-cxr-types.scm,v 1.6 1999/06/23 16:57:22 leavens Exp $
;;; Generate types for built-in Scheme procedures cxr (e.g., car, cdr, cadr)
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.

(load-quietly-from-lib "type-check-polytype.scm")

(define tc:compose-cxr-types
  ;; TYPE: (-> (external-type-expr external-type-expr) external-type-expr)
  (lambda (left right)
    ;; REQUIRES: left and right are and-external-type-expr
    (tc:make-and-external-type-expr
     (append (tc:external-type-expr->conjoined-types left)
	     (tc:external-type-expr->conjoined-types right)))))

(define tc:compose-arrow-tc:types
  ;; TYPE: (-> (external-type-expr external-type-expr)
  ;;           (list external-type-expr))
  (lambda (left-type right-type)
    ;; REQUIRES: both left-type and right-type are one-argument function types
    ;; ENSURES: result is a list containing a single
    ;;          function-external-type-expr, if the two types can be composed,
    ;;          otherwise an empty list.
    (let* ((left-result-type
	    (tc:type-untranslate
	     (tc:external-type-expr->function-result-type left-type)))
	   (right-result-type
	    (tc:type-untranslate
	     (tc:external-type-expr->function-result-type right-type)))
	   (right-arg-type
	    (tc:type-untranslate
	     (car (tc:external-type-expr->function-arg-types right-type))))
	   (left-arg-type 
	    (tc:type-untranslate
	     (car (tc:external-type-expr->function-arg-types left-type)))))
      (let ((e (tc:unify right-result-type left-arg-type (tc:env-empty) '())))
	;  	  (displayln "tc:compose-arrow-type got env " e)
	(if (tc:unification-error? e)
	    '()
	    (list (tc:type-translate
		   (tc:subst-bindings e
				      (tc:make-function-type-expr
				       (list right-arg-type)
				       left-result-type)))))))))

(define *tc:cxr-list*
  ;; (list (pair (symbol (list symbol))))
  '((caar car car)
    (cadr car cdr)
    (cdar cdr car)
    (cddr cdr cdr)

    (caaar car car car)
    (caadr car car cdr)
    (cadar car cdr car)
    (caddr car cdr cdr)
    (cdaar cdr car car)
    (cdadr cdr car cdr)
    (cddar cdr cdr car)
    (cdddr cdr cdr cdr)

    (caaaar car car car car)
    (caaadr car car car cdr)
    (caadar car car cdr car)
    (caaddr car car cdr cdr)
    (cadaar car cdr car car)
    (cadadr car cdr car cdr)
    (caddar car cdr cdr car)
    (cadddr car cdr cdr cdr)
    (cdaaar cdr car car car)
    (cdaadr cdr car car cdr)
    (cdadar cdr car cdr car)
    (cdaddr cdr car cdr cdr)
    (cddaar cdr cdr car car)
    (cddadr cdr cdr car cdr)
    (cdddar cdr cdr cdr car)
    (cddddr cdr cdr cdr cdr)
    ))

(define *tc:cxr-external-type-env*
  ;; TYPE: (env symbol external-type-expr)
  (letrec ((car-type (tc:env-value 'car *tc:pre-global-var-external-type-env*))
	   (cdr-type (tc:env-value 'cdr *tc:pre-global-var-external-type-env*))
	   (name->external-type-expr
	    ;; TYPE: (-> (symbol) external-type-expr)
 	    (lambda (name)
	      (if (eq? name 'car) car-type cdr-type)))
	   (names->external-type-expr
	    ;; TYPE: (-> ((list symbol)) external-type-expr)
	    (lambda (names-list)
	      (if (null? (cdr names-list))
		  (name->external-type-expr (car names-list))
		  (tc:compose-cxr-types
		   (name->external-type-expr (car names-list))
		   (names->external-type-expr (cdr names-list)))))))
    (tc:env-map
     (lambda (composition)
       (let ((name (car composition))
	     (recipe (cdr composition)))
	 (cons
	  name
	  (names->external-type-expr recipe))))
     (tc:assoc-list->env *tc:cxr-list*))))

