;;; $Id: type-check-external-type-expr.scm,v 1.2 1999/06/23 16:57:22 leavens Exp $
;;;
;;; External types, reported to and gotten from the user:
;;;
;;;   <type> ::= <external-variable>
;;;            | <basic-type>
;;;            | <function-type>
;;;            | <and-type>
;;;            | <applied>
;;;   <basic-type> ::= any <symbol> that is not a <variable>
;;;   <function-type> ::= (-> ( {<type>}* ) <type>)     ;; parens around args
;;;                     | (-> ( {<type>}* <type> ... ) <type>)
;;;   <and-type> ::= (and {<type>}+)
;;;   <applied> ::= (<symbol> {<type>}+)
;;;   <external-variable> ::= a single upper case letter
;;;                           or ? followed by id characters

(load-quietly-from-lib "type-check-variable.scm")

(define tc:external-type-expr?
  ;; TYPE: (-> (datum) boolean)
  (lambda (d)
    (or (tc:external-type-expr-variable? d)
	(tc:external-type-expr-basic-type? d)
	(and (tc:external-type-expr-function-type? d)
	     (tc:all tc:external-type-expr? (cdr d)))
	(and (tc:external-type-expr-and-type? d)
	     (tc:all tc:external-type-expr? (cdr d)))
	(and (tc:external-type-expr-applied-type? d)
	     (tc:all tc:external-type-expr? (cdr d))))))

(define tc:external-type-expr-variable?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (or (tc:variable? d)
	(and (symbol? d) (= 1 (string-length (symbol->string d)))))))

(define tc:external-type-expr-basic-type?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (and (symbol? d)
	 (not (tc:external-type-expr-variable? d)))))

(define tc:external-type-expr-poof?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (eq? d 'poof)))

(define tc:external-type-expr-datum?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (eq? d 'datum)))

(define tc:external-type-expr-dots?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (eq? d '...)))

(define tc:external-type-expr-function-type?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (and (list? d)
	 (= (length d) 3)
	 (eq? (car d) '->)
	 (list? (cadr d)))))

(define tc:external-type-expr-varargs-function-type?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (and (tc:external-type-expr-function-type? d)
	 (tc:member? '... (cadr d)))))

(define tc:external-type-expr-and-type?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (and (list? d)
	 (>= (length d) 2)
	 (eq? (car d) 'and))))

(define tc:external-type-expr-applied-type?
  ;; TYPE: (-> (external-type-expr) boolean)
  (lambda (d)
    (and (list? d)
	 (>= (length d) 2)
	 (symbol? (car d))
	 (not (eq? (car d) '->))
	 (not (eq? (car d) 'and))
	 (not (tc:variable? (car d))))))

(define tc:make-variable-external-type-expr
  ;; TYPE: (-> (variable) external-type-expr)
  (lambda (v) v))

(define tc:make-basic-external-type-expr
  ;; TYPE: (-> (symbol) external-type-expr)
  (lambda (s) s))

(define tc:make-function-external-type-expr
  ;; TYPE: (-> ((list external-type-expr) external-type-expr) external-type-expr)
  (lambda (arg-types result-type)
    (list '-> arg-types result-type)))

(define tc:make-and-external-type-expr
  ;; TYPE: (-> ((list external-type-expr)) external-type-expr)
  (lambda (arg-types)
    (cons 'and arg-types)))

(define tc:make-applied-external-type-expr
  ;; TYPE: (-> (symbol (list external-type-expr)) external-type-expr)
  cons)

(define tc:external-type-expr->variable
  ;; TYPE: (-> (external-type-expr) variable)
  (lambda (t)
    (if (tc:external-type-expr-variable? t)
	t
	(error "tc:external-type-expr->variable called with non-variable argument "
	       t))))

(define tc:external-type-expr->basic-type
  ;; TYPE: (-> (external-type-expr) symbol)
  (lambda (t)
    (if (tc:external-type-expr-basic-type? t)
	t
	(error "tc:external-type-expr->basic-type called with non-basic-type argument "
	       t))))

(define tc:external-type-expr->function-arg-types
  ;; TYPE: (-> (external-type-expr) (list external-type-expr))
  (lambda (t)
    (if (tc:external-type-expr-function-type? t)
	(cadr t)
	(error
	 "tc:external-type-expr->function-arg-types called with non-function type argument "
	 t))))

(define tc:external-type-expr->function-result-type
  ;; TYPE: (-> (external-type-expr) external-type-expr)
  (lambda (t)
    (if (tc:external-type-expr-function-type? t)
	(caddr t)
	(error
	 "tc:external-type-expr->function-result-type called with non-function type argument "
	 t))))

(define tc:external-type-expr->conjoined-types
  ;; TYPE: (-> (external-type-expr) (list external-type-expr))
  (lambda (t)
    (if (tc:external-type-expr-and-type? t)
	(cdr t)
	(error
	 "tc:external-type-expr->conjoined-types called with non-and-type argument "
	 t))))

(define tc:external-type-expr->operator
  ;; TYPE: (-> (external-type-expr) symbol)
  (lambda (t)
    (if (tc:external-type-expr-applied-type? t)
	(car t)
	(error
	 "tc:external-type-expr->operator called with non-applied type argument "
	 t))))

(define tc:external-type-expr->operands
  ;; TYPE: (-> (external-type-expr) (list external-type-expr))
  (lambda (t)
    (if (tc:external-type-expr-applied-type? t)
	(cdr t)
	(error
	 "tc:external-type-expr->operands called with non-applied type argument "
	 t))))

(define tc:parse-external-type-expr
  ;; TYPE: (-> (datum) external-type-expr)
  (lambda (d)
    (cond
    ((tc:external-type-expr-variable? d)
     (tc:make-variable-external-type-expr d))
    ((tc:external-type-expr-basic-type? d)
     (tc:make-basic-external-type-expr d))
    ((tc:external-type-expr-function-type? d)
     (tc:make-function-external-type-expr
      (map tc:parse-external-type-expr (cadr d))
      (tc:parse-external-type-expr (caddr d))))
    ((tc:external-type-expr-and-type? d)
     (tc:make-and-external-type-expr
      (map tc:parse-external-type-expr (cdr d))))
    ((tc:external-type-expr-applied-type? d)
     (tc:make-applied-external-type-expr
      (car d)
      (map tc:parse-external-type-expr (cdr d))))
    (else (error "tc:parse-external-type-expr syntax error in " d)))))
    
(define tc:occurs-in-external-type-expr?
  ;; TYPE: (-> (external-type-expr external-type-expr) boolean)
  (lambda (typ1 typ2)
    ;; EFFECT: determines if typ1 occurs anywhere in typ2
    (or
     (equal? typ1 typ2)
     (and (tc:external-type-expr-function-type? typ2)
	  (or (tc:some (lambda (t) (tc:occurs-in-external-type-expr? typ1 t))
		       (tc:external-type-expr->function-arg-types typ2))
	      (tc:occurs-in-external-type-expr?
	       typ1
	       (tc:external-type-expr->function-result-type typ2))))
     (and (tc:external-type-expr-and-type? typ2)
	  (tc:some (lambda (t) (tc:occurs-in-external-type-expr? typ1 t))
		   (tc:external-type-expr->conjoined-types typ2)))
     (and (tc:external-type-expr-applied-type? typ2)
	  (or (equal? typ1 (tc:external-type-expr->operator typ2))
	      (tc:some (lambda (t) (tc:occurs-in-external-type-expr? typ1 t))
		       (tc:external-type-expr->operands typ2)))))))
